## Copyright (C) 2005-2011 Nicolo' Giorgetti
##
## This file is part of Octave.
##
## Octave is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or (at
## your option) any later version.
##
## Octave is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, see
## <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn {Function File} {[@var{xopt}, @var{fmin}, @var{status}, @var{extra}] =} glpk (@var{c}, @var{A}, @var{b}, @var{lb}, @var{ub}, @var{ctype}, @var{vartype}, @var{sense}, @var{param})
## Solve a linear program using the GNU @sc{glpk} library.  Given three
## arguments, @code{glpk} solves the following standard LP:
## @tex
## $$
##   \min_x C^T x
## $$
## @end tex
## @ifnottex
##
## @example
## min C'*x
## @end example
##
## @end ifnottex
## subject to
## @tex
## $$
##   Ax = b \qquad x \geq 0
## $$
## @end tex
## @ifnottex
##
## @example
## @group
## A*x  = b
##   x >= 0
## @end group
## @end example
##
## @end ifnottex
## but may also solve problems of the form
## @tex
## $$
##   [ \min_x | \max_x ] C^T x
## $$
## @end tex
## @ifnottex
##
## @example
## [ min | max ] C'*x
## @end example
##
## @end ifnottex
## subject to
## @tex
## $$
##  Ax [ = | \leq | \geq ] b \qquad LB \leq x \leq UB
## $$
## @end tex
## @ifnottex
##
## @example
## @group
## A*x [ "=" | "<=" | ">=" ] b
##   x >= LB
##   x <= UB
## @end group
## @end example
##
## @end ifnottex
##
## Input arguments:
##
## @table @var
## @item c
## A column array containing the objective function coefficients.
##
## @item A
## A matrix containing the constraints coefficients.
##
## @item b
## A column array containing the right-hand side value for each constraint
## in the constraint matrix.
##
## @item lb
## An array containing the lower bound on each of the variables.  If
## @var{lb} is not supplied, the default lower bound for the variables is
## zero.
##
## @item ub
## An array containing the upper bound on each of the variables.  If
## @var{ub} is not supplied, the default upper bound is assumed to be
## infinite.
##
## @item ctype
## An array of characters containing the sense of each constraint in the
## constraint matrix.  Each element of the array may be one of the
## following values
## @table @asis
## @item "F"
## A free (unbounded) constraint (the constraint is ignored).
##
## @item "U"
## An inequality constraint with an upper bound (@code{A(i,:)*x <= b(i)}).
##
## @item "S"
## An equality constraint (@code{A(i,:)*x = b(i)}).
##
## @item "L"
## An inequality with a lower bound (@code{A(i,:)*x >= b(i)}).
##
## @item "D"
## An inequality constraint with both upper and lower bounds
## (@code{A(i,:)*x >= -b(i)} @emph{and} (@code{A(i,:)*x <= b(i)}).
## @end table
##
## @item vartype
## A column array containing the types of the variables.
## @table @asis
## @item "C"
## A continuous variable.
##
## @item "I"
## An integer variable.
##
## @item "B"
## A binary variable.
## @end table
##
## @item sense
## If @var{sense} is 1, the problem is a minimization.  If @var{sense} is
## -1, the problem is a maximization.  The default value is 1.
##
## @item param
## A structure containing the following parameters used to define the
## behavior of solver.  Missing elements in the structure take on default
## values, so you only need to set the elements that you wish to change
## from the default.
##
## Integer parameters:
##
## @table @code
## @item msglev (@w{@code{LPX_K_MSGLEV}}, default: 1)
## Level of messages output by solver routines:
## @table @asis
## @item 0
## No output.
##
## @item 1
## Error messages only.
##
## @item 2
## Normal output.
##
## @item 3
## Full output (includes informational messages).
## @end table
##
## @item scale (@w{@code{LPX_K_SCALE}}, default: 1)
## Scaling option:
## @table @asis
## @item 0
## Skip scaling.
##
## @item 1
## Perform geometric mean scaling.
##
## @item 2
## Perform equilibration scaling.
##
## @item 3
## Perform automatic scaling
##
## @item 4
## Round scale factors to nearest power of two
## @end table
##
## @item dual (@w{@code{LPX_K_DUAL}}, default: 0)
## Dual simplex option:
## @table @asis
## @item 0
## Do not use the dual simplex.
##
## @item 1
## If initial basic solution is dual feasible, use the dual simplex.
## 
## @item 2
## Use two phase dual simplex, or primal simplex if dual fails
## @end table
##
## @item price   (@w{@code{LPX_K_PRICE}}, default: 1)
## Pricing option (for both primal and dual simplex):
## @table @asis
## @item 0
## Textbook pricing.
##
## @item 1
## Steepest edge pricing.
## @end table
##
## @item r_test   (default: 0)
## @table @asis
## @item 0
## Standard (textbook).
## 
## @item 1
## Harris's two-pass ratio test.
## @end table
##
## @item round (@w{@code{LPX_K_ROUND}}, default: 0)
## Solution rounding option:
## @table @asis
## @item 0
## Report all primal and dual values "as is".
##
## @item 1
## Replace tiny primal and dual values by exact zero.
## @end table
##
## @item itlim (@w{@code{LPX_K_ITLIM}}, default: -1)
## Simplex iterations limit.  If this value is positive, it is decreased by
## one each time when one simplex iteration has been performed, and
## reaching zero value signals the solver to stop the search.  Negative
## value means no iterations limit.
##
## @item itcnt (@w{@code{LPX_K_OUTFRQ}}, default: 200)
## Output frequency, in iterations.  This parameter specifies how
## frequently the solver sends information about the solution to the
## standard output.
##
## @item branch (@w{@code{LPX_K_BRANCH}}, default: 2)
## Branching heuristic option (for MIP only):
## @table @asis
## @item 0
## Branch on the first variable.
##
## @item 1
## Branch on the last variable.
##
## @item 2
## Branch on the most fractional variable.
##
## @item 3
## Branch using a heuristic by Driebeck and Tomlin.
## @end table
##
## @item btrack (@w{@code{LPX_K_BTRACK}}, default: 2)
## Backtracking heuristic option (for MIP only):
## @table @asis
## @item 0
## Depth first search.
##
## @item 1
## Breadth first search.
##
## @item 2
## Best local bound.
##
## @item 3
## Backtrack using the best projection heuristic.
## @end table
##
## @item presol (@w{@code{LPX_K_PRESOL}}, default: 1)
## If this flag is set, the routine lpx_simplex solves the problem using
## the built-in LP presolver.  Otherwise the LP presolver is not used.
##
## @item lpsolver (default: 1)
## Select which solver to use.  If the problem is a MIP problem this flag
## will be ignored.
## @table @asis
## @item 1
## Revised simplex method.
##
## @item 2
## Interior point method.
##
## @item 3
## Simplex method with exact arithmetic
## @end table
##
## @item pprocess (default: 2)
## Pre-processing technique option (for MIP only).
## @table @asis
## @item 0
## Disable preprocessing.
##
## @item 1
## Perform preprocessing for root only.
##
## @item 2
## Perform preprocessing for all levels.
## @end table
##
## @item usecuts (default: 1)
## Adds cutting planes in order to improve the LP relaxation before
## applying the branch and bound method (for MIP only)
## @table @asis
## @item 0
## Do not use cuts.
##
## @item 1
## Use Gomoy's mixed integer cuts.
##
## @item 2
## Use mixed integer rounding cuts.
##
## @item 3
## Use mixed cover cuts.
##
## @item 4
## Use clique cuts.
##
## @item 5
## Use all cuts.
## @end table
##
## @item binarize (default: 0)
## Binarization option (for MIP only and used only if the presolver is
## enabled).
## @table @asis
## @item 0
## Do not use binarization.
##
## @item 1
## Replace general integer variables with binary ones.
## @end table
##
## @item save (default: 0)
## If this parameter is nonzero, a copy of the problem is saved to file. 
## Desired filename and format can be specified using the string parameters
## described below. If these are not defined, the original problem is 
## saved in CPLEX LP format to the file @file{outpb.lp}.  
##
## @item mpsinfo (default: 1)
## If this is set, the interface writes to file several comment cards,
## which contains some information about the problem. Otherwise the routine
## writes no comment cards.
##
## @item mpsobj (default: 2)
## This parameter tells the routing how to output the objective function
## row. 
## @table @asis
## @item 0
## Never output objective function row.
##
## @item 1
## Always output objective function row.
##
## @item 2
## Output objective function row if the problem has no free rows.
## @end table
##
## @item mpsorig (default: 0)
## If this is set, the routine uses the original symbolic names of rows
## and columns. Otherwise the routine generates plain names using ordinal
## numbers of rows and columns.
##
## @item mpswide (default: 1)
## If this is set, the routine uses all data fields. Otherwise the routine
## keeps fields 5 and 6 empty.
##
## @item mpsfree (default: 0)
## If this is set, the routine omits column and vector names every time
## when possible (free style). Otherwise the routine never omits these
## names (pedantic style).
## @end table
##
## Real parameters:
##
## @table @code
## @item relax (@w{@code{LPX_K_RELAX}}, default: 0.07)
## Relaxation parameter used in the ratio test.  If it is zero, the textbook
## ratio test is used.  If it is non-zero (should be positive), Harris'
## two-pass ratio test is used.  In the latter case on the first pass of the
## ratio test basic variables (in the case of primal simplex) or reduced
## costs of non-basic variables (in the case of dual simplex) are allowed
## to slightly violate their bounds, but not more than
## @code{relax*tolbnd} or @code{relax*toldj (thus, @code{relax} is a
## percentage of @code{tolbnd} or @code{toldj}}.
##
## @item tolbnd (@w{@code{LPX_K_TOLBND}}, default: 10e-7)
## Relative tolerance used to check if the current basic solution is primal
## feasible.  It is not recommended that you change this parameter unless you
## have a detailed understanding of its purpose.
##
## @item toldj (@w{@code{LPX_K_TOLDJ}}, default: 10e-7)
## Absolute tolerance used to check if the current basic solution is dual
## feasible.  It is not recommended that you change this parameter unless you
## have a detailed understanding of its purpose.
##
## @item tolpiv (@w{@code{LPX_K_TOLPIV}}, default: 10e-9)
## Relative tolerance used to choose eligible pivotal elements of the
## simplex table.  It is not recommended that you change this parameter unless
## you have a detailed understanding of its purpose.
##
## @item objll (@w{@code{LPX_K_OBJLL}}, default: -DBL_MAX)
## Lower limit of the objective function.  If on the phase II the objective
## function reaches this limit and continues decreasing, the solver stops
## the search.  This parameter is used in the dual simplex method only.
##
## @item objul (@w{@code{LPX_K_OBJUL}}, default: +DBL_MAX)
## Upper limit of the objective function.  If on the phase II the objective
## function reaches this limit and continues increasing, the solver stops
## the search.  This parameter is used in the dual simplex only.
##
## @item tmlim (@w{@code{LPX_K_TMLIM}}, default: -1.0)
## Searching time limit, in seconds.  If this value is positive, it is
## decreased each time when one simplex iteration has been performed by the
## amount of time spent for the iteration, and reaching zero value signals
## the solver to stop the search.  Negative value means no time limit.
##
## @item outdly (@w{@code{LPX_K_OUTDLY}}, default: 0.0)
## Output delay, in seconds.  This parameter specifies how long the solver
## should delay sending information about the solution to the standard
## output.  Non-positive value means no delay.
##
## @item tolint (@w{@code{LPX_K_TOLINT}}, default: 10e-5)
## Relative tolerance used to check if the current basic solution is integer
## feasible.  It is not recommended that you change this parameter unless
## you have a detailed understanding of its purpose.
##
## @item tolobj (@w{@code{LPX_K_TOLOBJ}}, default: 10e-7)
## Relative tolerance used to check if the value of the objective function
## is not better than in the best known integer feasible solution.  It is
## not recommended that you change this parameter unless you have a
## detailed understanding of its purpose.
##
## @item mipgap (@w{@code{LPX_K_MIPGAPJ}}, default: 0.0)
## The relative MIP gap tolerance. Inf the relative mip gap for currently
## known best integer feasible solution falls below this tolerance, the
## solver terminates the search. This allows obtaining suboptimal integer
## feasible solutions if solving the problem to optimality takes too long.
## @end table
##
## String Parameters:
## @table @code
## @item savefilename (default: @file{outpb})
## The parameter allows saving the original problem in a file of desired
## filename.
## @item savefiletype (default: @file{cplex})
## The parameter allows to specify the format type used to save the problem.
## @table @asis
## @item @file{fixedmps}
## Fixed MPS format.
##
## @item @file{freemps}
## Free MPS format.
##
## @item @file{cplex}
## CPLEX LP format.
##
## @item @file{plain}
## Plain text.
## @end table
## @end table
## @end table
##
## Output values:
##
## @table @var
## @item xopt
## The optimizer (the value of the decision variables at the optimum).
##
## @item fopt
## The optimum value of the objective function.
##
## @item status
## Status of the optimization.
##
## @table @asis
## @item 1
## Solution is undefined.
## @item 2
## Solution is feasible.
## @item 3
## Solution is infeasible.
## @item 4
## Problem has no feasible solution.
## @item 5
## Solution is optimal.
## @item 6
## Solution is unbounded.
## @end table
## @noindent
## If an error occurs, @var{status} will contain one of the following
## codes.
##
## Simplex method routines:
## @table @asis
## @item 101 (@w{@code{GLP_EBADB}})
## Unable to start the search, because the initial basis specifed in the 
## problem object is invalid|the number of basic (auxiliary and structural)
## variables is not the same as the number of rows in the problem object.
## @item 102 (@w{@code{GLP_ESING}})
## Unable to start the search, because the basis matrix corresponding
## to the initial basis is singular within the working precision.
## @item 103 (@w{@code{GLP_ECOND}})
## Unable to start the search, because the basis matrix corresponding
## to the initial basis is ill-conditioned, i.e. its condition number is 
## too large.
## @item 104 (@w{@code{GLP_EBOUND}})
## Unable to start the search, because some double-bounded (auxiliary or 
## structural) variables have incorrect bounds.
## @item 105 (@w{@code{GLP_EFAIL}})
## The search was prematurely terminated due to the solver failure.
## @item 106 (@w{@code{GLP_EOBJLL}})
## The search was prematurely terminated, because the objective function 
## being maximized has reached its lower limit and continues decreasing
## (the dual simplex only).
## @item 107 (@w{@code{GLP_EOBJUL}})
## The search was prematurely terminated, because the objective function 
## being minimized has reached its upper limit and continues increasing 
## (the dual simplex only).
## @item 108 (@w{@code{GLP_EITLIM}})
## The search was prematurely terminated, because the simplex iteration 
## limit has been exceeded.
## @item 109 (@w{@code{GLP_ETMLIM}})
## The search was prematurely terminated, because the time limit has been 
## exceeded.
## @item 110 (@w{@code{GLP_ENOPFS}})
## The LP problem instance has no primal feasible solution (only if the LP
## presolver is used).
## @item 111 (@w{@code{GLP_ENODFS}})
## The LP problem instance has no dual feasible solution (only if the LP 
## presolver is used).
## @end table
##
## Mixed integer programming routines:
## @table @asis
## @item 204 (@w{@code{GLP_EBOUND}})
## Unable to start the search, because some double-bounded
## variables have incorrect bounds or some integer variables
## have non-integer (fractional) bounds.
## @item 205 (@w{@code{GLP_EFAIL}})
## The search was prematurely terminated due to the solver failure.
## @item 209 (@w{@code{GLP_ETMLIM}})
## The search was prematurely terminated, because the time limit has been 
## exceeded.
## @item 210 (@w{@code{GLP_ENOPFS}})
## Unable to start the search, because LP relaxation of the
## MIP problem instance has no primal feasible solution.
## (This code may appear only if the presolver is enabled.)
## @item 211 (@w{@code{GLP_ENODFS}})
## Unable to start the search, because LP relaxation of the
## MIP problem instance has no dual feasible solution. In
## other word, this code means that if the LP relaxation has
## at least one primal feasible solution, its optimal solution is
## unbounded, so if the MIP problem has at least one integer
## feasible solution, its (integer) optimal solution is also unbounded.
## (This code may appear only if the presolver is enabled.)
## @item 212 (@w{@code{GLP_EROOT}})
## Unable to start the search, because optimal basis for initial
## LP relaxation is not provided. (This code may appear only
## if the presolver is disabled.)
## @item 214 (@w{@code{GLP_EMIPGAP}})
## The search was prematurely terminated, because the relative MIP gap 
## tolerance has been reached.
## @end table
##
## Interior point method routines:
## @table @asis
## @item 305 (@w{@code{GLP_EFAIL}})
## The problem has no rows/columns.
## @item 308 (@w{@code{GLP_EITLIM}})
## Iteration limit exceeded.
## @item 316 (@w{@code{GLP_ENOCVG}})
## Very slow convergence or divergence.
## @item 317 (@w{@code{GLP_EINSTAB}})
## Numerical instability on solving Newtonian system.
## @end table
##
## @item extra
## A data structure containing the following fields:
## @table @code
## @item lambda
## Dual variables.
##
## @item redcosts
## Reduced Costs.
##
## @item time
## Time (in seconds) used for solving LP/MIP problem.
##
## @item mem
## Memory (in bytes) used for solving LP/MIP problem (this is not
## available if the version of @sc{glpk} is 4.15 or later).
## @end table
## @end table
##
## Example:
##
## @example
## @group
## c = [10, 6, 4]';
## A = [ 1, 1, 1;
##      10, 4, 5;
##       2, 2, 6];
## b = [100, 600, 300]';
## lb = [0, 0, 0]';
## ub = [];
## ctype = "UUU";
## vartype = "CCC";
## s = -1;
##
## param.msglev = 1;
## param.itlim = 100;
##
## [xmin, fmin, status, extra] = ...
##    glpk (c, A, b, lb, ub, ctype, vartype, s, param);
## @end group
## @end example
## @end deftypefn

## Author: Nicolo' Giorgetti <giorgetti@dii.unisi.it>
## Adapted-by: jwe

function [xopt, fmin, status, extra] = glpk (c, A, b, lb, ub, ctype, vartype, sense, param)

  ## If there is no input output the version and syntax
  if (nargin < 3 || nargin > 9)
    print_usage ();
    return;
  endif

  if (all (size (c) > 1) || iscomplex (c) || ischar (c))
    error ("glpk: c must be a real vector");
    return;
  endif
  nx = length (c);
  ## Force column vector.
  c = c(:);

  ## 2) Matrix constraint

  if (isempty (A))
    error ("glpk: A cannot be an empty matrix");
    return;
  endif
  [nc, nxa] = size(A);
  if (! isreal (A) || nxa != nx)
    error ("glpk: A must be a real valued %d by %d matrix", nc, nx);
    return;
  endif

  ## 3) RHS

  if (isempty (b))
    error ("glpk: B cannot be an empty vector");
    return;
  endif
  if (! isreal (b) || length (b) != nc)
    error ("glpk: B must be a real valued %d by 1 vector", nc);
    return;
  endif

  ## 4) Vector with the lower bound of each variable

  if (nargin > 3)
    if (isempty (lb))
      lb = zeros (nx, 1);
    elseif (! isreal (lb) || all (size (lb) > 1) || length (lb) != nx)
      error ("glpk: LB must be a real valued %d by 1 column vector", nx);
      return;
    endif
  else
    lb = zeros (nx, 1);
  endif

  ## 5) Vector with the upper bound of each variable

  if (nargin > 4)
    if (isempty (ub))
      ub = Inf (nx, 1);
    elseif (! isreal (ub) || all (size (ub) > 1) || length (ub) != nx)
      error ("glpk: UB must be a real valued %d by 1 column vector", nx);
      return;
    endif
  else
    ub = Inf (nx, 1);
  endif

  ## 6) Sense of each constraint

  if (nargin > 5)
    if (isempty (ctype))
      ctype = repmat ("S", nc, 1);
    elseif (! ischar (ctype) || all (size (ctype) > 1) || length (ctype) != nc)
      error ("glpk: CTYPE must be a char valued vector of length %d", nc);
      return;
    elseif (! all (ctype == "F" | ctype == "U" | ctype == "S"
                   | ctype == "L" | ctype == "D"))
      error ("glpk: CTYPE must contain only F, U, S, L, or D");
      return;
    endif
  else
    ctype = repmat ("S", nc, 1);
  endif

  ## 7) Vector with the type of variables

  if (nargin > 6)
    if (isempty (vartype))
      vartype = repmat ("C", nx, 1);
    elseif (! ischar (vartype) || all (size (vartype) > 1)
            || length (vartype) != nx)
      error ("glpk: VARTYPE must be a char valued vector of length %d", nx);
      return;
    elseif (! all (vartype == "C" | vartype == "I" | vartype == "B"))
      error ("glpk: VARTYPE must contain only C, I or B");
      return;
    endif
  else
    ## As default we consider continuous vars
    vartype = repmat ("C", nx, 1);
  endif

  ## 8) Sense of optimization

  if (nargin > 7)
    if (isempty (sense))
      sense = 1;
    elseif (ischar (sense) || all (size (sense) > 1) || ! isreal (sense))
      error ("glpk: SENSE must be an integer value");
    elseif (sense >= 0)
      sense = 1;
    else
      sense = -1;
    endif
  else
    sense = 1;
  endif

  ## 9) Parameters vector

  if (nargin > 8)
    if (! isstruct (param))
      error ("glpk: PARAM must be a structure");
      return;
    endif
  else
    param = struct ();
  endif

  [xopt, fmin, status, extra] = ...
    __glpk__ (c, A, b, lb, ub, ctype, vartype, sense, param);

endfunction
