/*

Copyright (C) 2011 Tommaso Balercia
Copyright (C) 2005-2011 Nicolo' Giorgetti

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cfloat>
#include <csetjmp>
#include <ctime>
#include <cstring>
#include <string>

#include "lo-ieee.h"

#include "defun-dld.h"
#include "error.h"
#include "gripes.h"
#include "oct-map.h"
#include "oct-obj.h"
#include "pager.h"

#if defined (HAVE_GLPK)

extern "C" {
#include <glpk.h>
}

#define NIntP 21
#define NRealP 11

// control parameters not defined in glpk.h
#define LPX_K_PREPROCESS    401   /* preprocessing  */
#define LPX_K_RATIO_TEST    402   /* ratio test */


// Integer Param Defaluts
int glpIntParam[NIntP] = {
    0,
    1,
    0,
    1,
    0,
    INT_MAX,
    INT_MAX,
    200,
    1,
    2,
    0,
    1,
    0,
    0,
    3,
    2,
    1,
    0,
    2,
    0,
    1
};

// Integer Param Names
int IParam[NIntP] = {
    LPX_K_MSGLEV,
    LPX_K_SCALE,
    LPX_K_DUAL,
    LPX_K_PRICE,
    LPX_K_ROUND,
    LPX_K_ITLIM,
    LPX_K_ITCNT,
    LPX_K_OUTFRQ,
    LPX_K_MPSINFO,
    LPX_K_MPSOBJ,
    LPX_K_MPSORIG,
    LPX_K_MPSWIDE,
    LPX_K_MPSFREE,
    LPX_K_MPSSKIP,
    LPX_K_BRANCH,
    LPX_K_BTRACK,
    LPX_K_PRESOL,
    LPX_K_USECUTS,
    LPX_K_PREPROCESS,
    LPX_K_BINARIZE,
    LPX_K_RATIO_TEST
};

//Real Param Values
double glpRealParam[NRealP] = {
    0.07,
    1e-7,
    1e-7,
    1e-10,
    -DBL_MAX,
    DBL_MAX,
    INT_MAX,
    0.0,
    1e-5,
    1e-7,
    0.0
};

//Real Param Names
int RParam[NRealP] = {
    LPX_K_RELAX,
    LPX_K_TOLBND,
    LPX_K_TOLDJ,
    LPX_K_TOLPIV,
    LPX_K_OBJLL,
    LPX_K_OBJUL,
    LPX_K_TMLIM,
    LPX_K_OUTDLY,
    LPX_K_TOLINT,
    LPX_K_TOLOBJ,
    LPX_K_MIPGAP
};

static jmp_buf mark;  /*-- Address for long jump */

static int glpk_print_hook (void *info, const char *msg)
{
    int msglev = *reinterpret_cast<int*>(info);

    if(msglev > 0)
    {
        message(0, "%s", msg);
    }
    return 1;
}

//
int glpk (int sense, int n, int m, const double *c, int nz, const int *rn, const int *cn,
          const double *a, const double *b, const char *ctype, const int *freeLB, const double *lb,
          const int *freeUB, const double *ub, const int *vartype, int isMIP, int lpsolver,
          int save_pb, const char * filename, const char * filetype, double *xmin, double *fmin, double *status,
          double *lambda, double *redcosts, double *time, double *mem)
{
    int typx = 0;
    int method;

    clock_t t_start = clock();

    //Redirect standard output
    // Unfortunately at the moment parts of GLPK still print out messages
    // without checking the specified options (e.g. the conflict graph part). 
    // Passing msglev directly to the callback function is a workaround 
    // for this problem, when the user desires no output.
    int msglev=glpIntParam[0];
    glp_term_hook (glpk_print_hook, reinterpret_cast<void*>(&msglev));

    //-- Create an empty LP/MILP object
    glp_prob *lp = glp_create_prob ();

    //-- Set the sense of optimization
    if (sense == 1)
        glp_set_obj_dir (lp, GLP_MIN);
    else
        glp_set_obj_dir (lp, GLP_MAX);

    //-- Define the number of unknowns and their domains.
    glp_add_cols (lp, n);
    for (int i = 0; i < n; i++)
    {
        //-- Define type of the structural variables
        if (! freeLB[i] && ! freeUB[i]) {
            if ( lb[i] == ub[i] )
                glp_set_col_bnds (lp, i+1, GLP_FX, lb[i], ub[i]);
            else
                glp_set_col_bnds (lp, i+1, GLP_DB, lb[i], ub[i]);
        }
        else
        {
            if (! freeLB[i] && freeUB[i])
                glp_set_col_bnds (lp, i+1, GLP_LO, lb[i], ub[i]);
            else
            {
                if (freeLB[i] && ! freeUB[i])
                    glp_set_col_bnds (lp, i+1, GLP_UP, lb[i], ub[i]);
                else
                    glp_set_col_bnds (lp, i+1, GLP_FR, lb[i], ub[i]);
            }
        }

        // -- Set the objective coefficient of the corresponding
        // -- structural variable. No constant term is assumed.
        glp_set_obj_coef(lp,i+1,c[i]);

        if (isMIP)
            glp_set_col_kind (lp, i+1, vartype[i]);
    }

    glp_add_rows (lp, m);

    for (int i = 0; i < m; i++)
    {
        /*  If the i-th row has no lower bound (types F,U), the
            correspondent parameter will be ignored.
            If the i-th row has no upper bound (types F,L), the correspondent
            parameter will be ignored.
            If the i-th row is of S type, the i-th LB is used, but
            the i-th UB is ignored.
        */

        switch (ctype[i])
        {
        case 'F':
            typx = GLP_FR;
            break;
            // upper bound
        case 'U':
            typx = GLP_UP;
            break;
            // lower bound
        case 'L':
            typx = GLP_LO;
            break;
            // fixed constraint
        case 'S':
            typx = GLP_FX;
            break;
            // double-bounded variable
        case 'D':
            typx = GLP_DB;
            break;
        }

        if ( typx == GLP_DB && -b[i] < b[i]) {
            glp_set_row_bnds (lp, i+1, typx, -b[i], b[i]);
        }
        else if(typx == GLP_DB && -b[i] == b[i]) {
            glp_set_row_bnds (lp, i+1, GLP_FX, b[i], b[i]);
        }
        else {
            glp_set_row_bnds (lp, i+1, typx, b[i], b[i]);
        }

    }
    // Load constraint matrix A
    glp_load_matrix (lp, nz, rn, cn, a);

    // Save problem
    if (save_pb) {
        if (!strcmp(filetype,"cplex")) {
            if (glp_write_lp (lp, NULL, filename) != 0) {
                error("glpk: unable to write the problem to file");
                longjmp (mark, -1);
            }
        } else {
            if (!strcmp(filetype,"fixedmps")) {
                if (glp_write_mps (lp, GLP_MPS_DECK, NULL, filename) != 0) {
                    error("glpk: unable to write the problem to file");
                    longjmp (mark, -1);
                }
            } else {
                if (!strcmp(filetype,"freemps")) {
                    if (glp_write_mps (lp, GLP_MPS_FILE, NULL, filename) != 0) {
                        error("glpk: unable to write the problem to file");
                        longjmp (mark, -1);
                    }
                } else { // plain text
                    if (lpx_print_prob (lp, filename) != 0) {
                        error("glpk: unable to write the problem to file");
                        longjmp (mark, -1);
                    }
                }
            }
        }
    }
    //-- scale the problem data (if required)
    if (! glpIntParam[16] || lpsolver != 1) {
        switch ( glpIntParam[1] ) {
        case ( 0 ):
            glp_scale_prob( lp, GLP_SF_SKIP );
            break;
        case ( 1 ):
            glp_scale_prob( lp, GLP_SF_GM );
            break;
        case ( 2 ):
            glp_scale_prob( lp, GLP_SF_EQ );
            break;
        case ( 3 ):
            glp_scale_prob( lp, GLP_SF_AUTO  );
            break;
        case ( 4 ):
            glp_scale_prob( lp, GLP_SF_2N );
            break;
        default :
            error("glpk: unrecognized scaling option");
            longjmp (mark, -1);
        }
    }

    //-- build advanced initial basis (if required)
    if (lpsolver == 1 && ! glpIntParam[16])
        glp_adv_basis (lp, 0);

    glp_smcp sParam;
    glp_init_smcp(&sParam);

    //-- set control parameters for simplex/exact method
    if (lpsolver == 1 || lpsolver == 3) {
        //remap of control parameters for simplex method
        sParam.msg_lev=glpIntParam[0];	// message level

        // simplex method: primal/dual
        switch ( glpIntParam[2] ) {
        case 0:
            sParam.meth=GLP_PRIMAL;
            break;
        case 1:
            sParam.meth=GLP_DUAL;
            break;
        case 2:
            sParam.meth=GLP_DUALP;
            break;
        default:
            error("glpk: unrecognized primal/dual method");
            longjmp (mark, -1);
        }

        // pricing technique
        if (glpIntParam[3]==0) sParam.pricing=GLP_PT_STD;
        else sParam.pricing=GLP_PT_PSE;

        // ratio test
        if (glpIntParam[20]==0) sParam.r_test = GLP_RT_STD;
        else sParam.r_test=GLP_RT_HAR;

        //tollerances
        sParam.tol_bnd=glpRealParam[1];	// primal feasible tollerance
        sParam.tol_dj=glpRealParam[2];	// dual feasible tollerance
        sParam.tol_piv=glpRealParam[3];	// pivot tollerance
        sParam.obj_ll=glpRealParam[4];	// lower limit
        sParam.obj_ul=glpRealParam[5];	// upper limit

        // iteration limit
        if (glpIntParam[5]==-1) sParam.it_lim=INT_MAX;
        else sParam.it_lim=glpIntParam[5];

        // time limit
        if (glpRealParam[6]==-1) sParam.tm_lim=INT_MAX;
        else sParam.tm_lim=static_cast<int>(glpRealParam[6]);
        sParam.out_frq=glpIntParam[7];	// output frequency
        sParam.out_dly=static_cast<int>(glpRealParam[7]);	// output delay
        // presolver
        if (glpIntParam[16]) sParam.presolve=GLP_ON;
        else sParam.presolve=GLP_OFF;
    } else {
        for(int i = 0; i < NIntP; i++) {
            // skip assinging ratio test or
            if ( i == 18 || i == 20) continue;
            lpx_set_int_parm (lp, IParam[i], glpIntParam[i]);
        }

        for (int i = 0; i < NRealP; i++) {
            lpx_set_real_parm (lp, RParam[i], glpRealParam[i]);
        }
    }

    //set MIP params if MIP....
    glp_iocp iParam;
    glp_init_iocp(&iParam);

    if ( isMIP ) {
        method = 'I';

        switch (glpIntParam[0]) { //message level
        case 0:
            iParam.msg_lev = GLP_MSG_OFF;
            break;
        case 1:
            iParam.msg_lev = GLP_MSG_ERR;
            break;
        case 2:
            iParam.msg_lev = GLP_MSG_ON;
            break;
        case 3:
            iParam.msg_lev = GLP_MSG_ALL;
            break;
        default:
            error("glpk: unrecognized msg_lev option");
        }
        switch (glpIntParam[14]) { //branching param
        case 0:
            iParam.br_tech = GLP_BR_FFV;
            break;
        case 1:
            iParam.br_tech = GLP_BR_LFV;
            break;
        case 2:
            iParam.br_tech = GLP_BR_MFV;
            break;
        case 3:
            iParam.br_tech = GLP_BR_DTH;
            break;
        default:
            error("glpk: unrecognized branch option");
        }
        switch (glpIntParam[15]) { //backtracking heuristic
        case 0:
            iParam.bt_tech = GLP_BT_DFS;
            break;
        case 1:
            iParam.bt_tech = GLP_BT_BFS;
            break;
        case 2:
            iParam.bt_tech = GLP_BT_BLB;
            break;
        case 3:
            iParam.bt_tech = GLP_BT_BPH;
            break;
        default:
            error("glpk: unrecognized backtrack option");
        }

        if (  glpRealParam[8] > 0.0 && glpRealParam[8] < 1.0 )
            iParam.tol_int = glpRealParam[8];  // absolute tolorence
        else
            error("glpk: tolint must be between 0 and 1");

        iParam.tol_obj = glpRealParam[9];  // relative tolarence
        iParam.mip_gap = glpRealParam[10]; // realative gap tolerance

        // set time limit for mip
        if ( glpRealParam[6] < 0.0 || glpRealParam[6] > 1e6 )
            iParam.tm_lim = INT_MAX;
        else
            iParam.tm_lim = static_cast<int>(1000.0 * glpRealParam[6] );

        // Choose Cutsets for mip
        // shut all cuts off, then start over....
        iParam.gmi_cuts = GLP_OFF;
        iParam.mir_cuts = GLP_OFF;
        iParam.cov_cuts = GLP_OFF;
        iParam.clq_cuts = GLP_OFF;

        switch( glpIntParam[17] ) {
        case 0:
            break;
        case 1:
            iParam.gmi_cuts = GLP_ON;
            break;
        case 2:
            iParam.mir_cuts = GLP_ON;
            break;
        case 3:
            iParam.cov_cuts = GLP_ON;
            break;
        case 4:
            iParam.clq_cuts = GLP_ON;
            break;
        case 5:
            iParam.clq_cuts = GLP_ON;
            iParam.gmi_cuts = GLP_ON;
            iParam.mir_cuts = GLP_ON;
            iParam.cov_cuts = GLP_ON;
            break;
        default:
            error("glpk: unrecognized cut option");
        }

        switch( glpIntParam[18] ) { // pre-processing for mip
        case 0:
            iParam.pp_tech = GLP_PP_NONE;
            break;
        case 1:
            iParam.pp_tech = GLP_PP_ROOT;
            break;
        case 2:
            iParam.pp_tech = GLP_PP_ALL;
            break;
        default:
            error("glpk: unrecognized pprocess option");
        }

        if (glpIntParam[16])  iParam.presolve=GLP_ON;
        else                  iParam.presolve=GLP_OFF;

        if (glpIntParam[19])  iParam.binarize = GLP_ON;
        else                  iParam.binarize = GLP_OFF;

    }
    else {
        /* Choose simplex method ('S')
        or interior point method ('T')
        or Exact method          ('E')
        to solve the problem  */
        switch (lpsolver) {
        case 1:
            method = 'S';
            break;
        case 2:
            method = 'T';
            break;
        case 3:
            method = 'E';
            break;
        default:
            error("glpk:  unrecognized lpsolver option");
            longjmp (mark, -1);
        }
    }

    // now run the problem...
    int errnum = 0;

    switch (method) {
    case 'I':
        errnum = glp_intopt( lp, &iParam );
        errnum += 200; //this is to avoid ambiguity in the return codes.
        break;

    case 'S':
        errnum = glp_simplex(lp, &sParam);
        errnum += 100; //this is to avoid ambiguity in the return codes.
        break;

    case 'T':
        errnum = glp_interior(lp, NULL );
        errnum += 300; //this is to avoid ambiguity in the return codes.
        break;

    case 'E':
        errnum = glp_exact(lp, &sParam);
        errnum += 100; //this is to avoid ambiguity in the return codes.
        break;

    default:  /*xassert (method != method); */
        error("glpk: unrecognized problem");
        longjmp (mark, -1);

    }

    if (errnum==100 || errnum==200 || errnum==300 || errnum==106 || errnum==107 || errnum==108 || errnum==109 || errnum==209 || errnum==214 || errnum==308) {
        // Get status and object value
        if (isMIP) {
            *status = glp_mip_status (lp);
            *fmin = glp_mip_obj_val (lp);
        }
        else {

            if (lpsolver == 1 || lpsolver == 3) {
                *status = glp_get_status (lp);
                *fmin = glp_get_obj_val (lp);
            }
            else {
                *status = glp_ipt_status (lp);
                *fmin = glp_ipt_obj_val (lp);
            }
        }

        // Get optimal solution (if exists)
        if (isMIP) {

            for (int i = 0; i < n; i++)
                xmin[i] = glp_mip_col_val (lp, i+1);
        }
        else {

            /* Primal values */
            for (int i = 0; i < n; i++) {

                if (lpsolver == 1 || lpsolver == 3)
                    xmin[i] = glp_get_col_prim (lp, i+1);
                else
                    xmin[i] = glp_ipt_col_prim (lp, i+1);
            }

            /* Dual values */
            for (int i = 0; i < m; i++) {

                if (lpsolver == 1 || lpsolver == 3)
                    lambda[i] = glp_get_row_dual (lp, i+1);
                else
                    lambda[i] = glp_ipt_row_dual (lp, i+1);
            }

            /* Reduced costs */
            for (int i = 0; i < glp_get_num_cols (lp); i++) {

                if (lpsolver == 1 || lpsolver == 3)
                    redcosts[i] = glp_get_col_dual (lp, i+1);
                else
                    redcosts[i] = glp_ipt_col_dual (lp, i+1);
            }

        }

        *time = (clock () - t_start) / CLOCKS_PER_SEC;

        glp_long tpeak;
        glp_mem_usage(NULL, NULL, NULL, &tpeak);
        *mem=static_cast<double>((4294967296.0 * tpeak.hi + tpeak.lo) / (1024));

        glp_delete_prob (lp);

        return 0;
    }
    else {
        // printf("errnum is %d\n", errnum);
    }

    glp_delete_prob (lp);

    /* this shouldn't be nessiary with glp_deleted_prob, but try it
    if we have weird behavior again... */
    glp_free_env();


    *status = errnum;

    return errnum;
}

#endif

#define OCTAVE_GLPK_GET_REAL_PARAM(NAME, IDX) \
  do \
    { \
      octave_value tmp = PARAM.getfield (NAME); \
 \
      if (tmp.is_defined ()) \
        { \
          if (! tmp.is_empty ()) \
            { \
              glpRealParam[IDX] = tmp.scalar_value (); \
 \
              if (error_state) \
                { \
                  error ("glpk: invalid value in PARAM." NAME); \
                  return retval; \
                } \
            } \
          else \
            { \
              error ("glpk: invalid value in PARAM." NAME); \
              return retval; \
            } \
        } \
    } \
  while (0)

#define OCTAVE_GLPK_GET_INT_PARAM(NAME, VAL) \
  do \
    { \
      octave_value tmp = PARAM.getfield (NAME); \
 \
      if (tmp.is_defined ()) \
        { \
          if (! tmp.is_empty ()) \
            { \
              VAL = tmp.int_value (); \
 \
              if (error_state) \
                { \
                  error ("glpk: invalid value in PARAM." NAME); \
                  return retval; \
                } \
            } \
          else \
            { \
              error ("glpk: invalid value in PARAM." NAME); \
              return retval; \
            } \
        } \
    } \
  while (0)

DEFUN_DLD (__glpk__, args, ,
           "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {[@var{values}] =} __glpk__ (@var{args})\n\
Undocumented internal function.\n\
@end deftypefn")
{
    // The list of values to return.  See the declaration in oct-obj.h
    octave_value_list retval;

#if defined (HAVE_GLPK)

    int nrhs = args.length ();

    if (nrhs != 9)
    {
        print_usage ();
        return retval;
    }

    //-- 1nd Input. A column array containing the objective function
    //--            coefficients.
    volatile int mrowsc = args(0).rows();

    Matrix C (args(0).matrix_value ());

    if (error_state)
    {
        error ("glpk: invalid value of c");
        return retval;
    }

    double *c = C.fortran_vec ();
    Array<int> rn;
    Array<int> cn;
    ColumnVector a;
    volatile int mrowsA;
    volatile int nz = 0;

    //-- 2nd Input. A matrix containing the constraints coefficients.
    // If matrix A is NOT a sparse matrix
    if (args(1).is_sparse_type ())
    {
        SparseMatrix A = args(1).sparse_matrix_value (); // get the sparse matrix

        if (error_state)
        {
            error ("glpk: invalid value of A");
            return retval;
        }

        mrowsA = A.rows ();
        octave_idx_type Anc = A.cols ();
        octave_idx_type Anz = A.nnz ();
        rn.resize (dim_vector (Anz+1, 1));
        cn.resize (dim_vector (Anz+1, 1));
        a.resize (Anz+1, 0.0);

        if (Anc != mrowsc)
        {
            error ("glpk: invalid value of A");
            return retval;
        }

        for (octave_idx_type j = 0; j < Anc; j++)
            for (octave_idx_type i = A.cidx(j); i < A.cidx(j+1); i++)
            {
                nz++;
                rn(nz) = A.ridx(i) + 1;
                cn(nz) = j + 1;
                a(nz) = A.data(i);
            }
    }
    else
    {
        Matrix A (args(1).matrix_value ()); // get the matrix

        if (error_state)
        {
            error ("glpk: invalid value of A");
            return retval;
        }

        mrowsA = A.rows ();
        rn.resize (dim_vector (mrowsA*mrowsc+1, 1));
        cn.resize (dim_vector (mrowsA*mrowsc+1, 1));
        a.resize (mrowsA*mrowsc+1, 0.0);

        for (int i = 0; i < mrowsA; i++)
        {
            for (int j = 0; j < mrowsc; j++)
            {
                if (A(i,j) != 0)
                {
                    nz++;
                    rn(nz) = i + 1;
                    cn(nz) = j + 1;
                    a(nz) = A(i,j);
                }
            }
        }

    }

    //-- 3rd Input. A column array containing the right-hand side value
    //               for each constraint in the constraint matrix.
    Matrix B (args(2).matrix_value ());

    if (error_state)
    {
        error ("glpk: invalid value of B");
        return retval;
    }

    double *b = B.fortran_vec ();

    //-- 4th Input. An array of length mrowsc containing the lower
    //--            bound on each of the variables.
    Matrix LB (args(3).matrix_value ());

    if (error_state || LB.length () < mrowsc)
    {
        error ("glpk: invalid value of LB");
        return retval;
    }

    double *lb = LB.fortran_vec ();

    //-- LB argument, default: Free
    Array<int> freeLB (dim_vector (mrowsc, 1));
    for (int i = 0; i < mrowsc; i++)
    {
        if (xisinf (lb[i]))
        {
            freeLB(i) = 1;
            lb[i] = -octave_Inf;
        }
        else
            freeLB(i) = 0;
    }

    //-- 5th Input. An array of at least length numcols containing the upper
    //--            bound on each of the variables.
    Matrix UB (args(4).matrix_value ());

    if (error_state || UB.length () < mrowsc)
    {
        error ("glpk: invalid value of UB");
        return retval;
    }

    double *ub = UB.fortran_vec ();

    Array<int> freeUB (dim_vector (mrowsc, 1));
    for (int i = 0; i < mrowsc; i++)
    {
        if (xisinf (ub[i]))
        {
            freeUB(i) = 1;
            ub[i] = octave_Inf;
        }
        else
            freeUB(i) = 0;
    }

    //-- 6th Input. A column array containing the sense of each constraint
    //--            in the constraint matrix.
    charMatrix CTYPE (args(5).char_matrix_value ());

    if (error_state)
    {
        error ("glpk: invalid value of CTYPE");
        return retval;
    }

    char *ctype = CTYPE.fortran_vec ();

    //-- 7th Input. A column array containing the types of the variables.
    charMatrix VTYPE (args(6).char_matrix_value ());

    if (error_state)
    {
        error ("glpk: invalid value of VARTYPE");
        return retval;
    }

    Array<int> vartype (dim_vector (mrowsc, 1));
    volatile int isMIP = 0;
    for (int i = 0; i < mrowsc ; i++)
    {
        switch(VTYPE(i,0)) {
        case 'I':
            vartype(i) = GLP_IV;
            isMIP = 1;
            break;
        case 'B':
            vartype(i) = GLP_BV;
            isMIP = 1;
            break;
        default:
            vartype(i) = GLP_CV;
            isMIP = 0;
        }
    }

    //-- 8th Input. Sense of optimization.
    volatile int sense;
    double SENSE = args(7).scalar_value ();

    if (error_state)
    {
        error ("glpk: invalid value of SENSE");
        return retval;
    }

    if (SENSE >= 0)
        sense = 1;
    else
        sense = -1;

    //-- 9th Input. A structure containing the control parameters.
    octave_scalar_map PARAM = args(8).scalar_map_value ();

    if (error_state)
    {
        error ("glpk: invalid value of PARAM");
        return retval;
    }

    //-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //-- Integer parameters
    //-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    //-- Level of messages output by the solver
    OCTAVE_GLPK_GET_INT_PARAM ("msglev", glpIntParam[0]);
    if (glpIntParam[0] < 0 || glpIntParam[0] > 3)
    {
        error ("glpk: the parameter \"msglev\" must be 0 (no output [default]) or 1 (error messages only) or 2 (normal output) or 3 (full output)");
        return retval;
    }

    //-- scaling option
    OCTAVE_GLPK_GET_INT_PARAM ("scale", glpIntParam[1]);
    if (glpIntParam[1] < 0 || glpIntParam[1] > 4)
    {
        error ("glpk: the parameter \"scale\" must be 0 (no scaling) or 1 (equilibration scaling [default]) or 2 (geometric mean scaling)");
        return retval;
    }

    //-- Dual dimplex option
    OCTAVE_GLPK_GET_INT_PARAM ("dual", glpIntParam[2]);
    if (glpIntParam[2] < 0 || glpIntParam[2] > 2)
    {
        error ("glpk: the parameters \"dual\" must be 0 (do NOT use dual simplex [default]) or 1 (use dual simplex)");
        return retval;
    }

    //-- Pricing option
    OCTAVE_GLPK_GET_INT_PARAM ("price", glpIntParam[3]);
    if (glpIntParam[3] < 0 || glpIntParam[3] > 1)
    {
        error ("glpk: the parameter \"price\" must be 0 (textbook pricing) or 1 (steepest edge pricing [default])");
        return retval;
    }

    //-- Ratio test option
    OCTAVE_GLPK_GET_INT_PARAM ("r_test", glpIntParam[20]);
    if (glpIntParam[20] < 0 || glpIntParam[20] > 1)
    {
        error ("glpk: the parameter \"r_test\" must be 0 (textbook [default]) or 1 (Harris's Two pass ratio test)");
        return retval;
    }

    //-- Solution rounding option
    OCTAVE_GLPK_GET_INT_PARAM ("round", glpIntParam[4]);
    if (glpIntParam[4] < 0 || glpIntParam[4] > 1)
    {
        error ("glpk: the parameter \"round\" must be 0 (report all primal and dual values [default]) or 1 (replace tiny primal and dual values by exact zero)");
        return retval;
    }

    //-- Simplex iterations limit
    OCTAVE_GLPK_GET_INT_PARAM ("itlim", glpIntParam[5]);

    //-- Simplex iterations count
    OCTAVE_GLPK_GET_INT_PARAM ("itcnt", glpIntParam[6]);

    //-- Output frequency, in iterations
    OCTAVE_GLPK_GET_INT_PARAM ("outfrq", glpIntParam[7]);

    //-- Branching heuristic option
    OCTAVE_GLPK_GET_INT_PARAM ("branch", glpIntParam[14]);
    if (glpIntParam[14] < 0 || glpIntParam[14] > 3)
    {
        error ("glpk: the parameter \"branch\" must be 0 (branch on first variable) or 1 (branch on last variable) or 2 (most fractional variable) or 3 (branch using a heuristic by Driebeck and Tomlin [default]");
        return retval;
    }

    //-- Backtracking heuristic option
    OCTAVE_GLPK_GET_INT_PARAM ("btrack", glpIntParam[15]);
    if (glpIntParam[15] < 0 || glpIntParam[15] > 3)
    {
        error ("glpk: the parameter \"btrack\" must be 0 (depth first search) or 1 (breadth first search) or 2 ( best local bound ) or 3 (backtrack using the best projection heuristic [default]");
        return retval;
    }

    //-- Presolver option
    OCTAVE_GLPK_GET_INT_PARAM ("presol", glpIntParam[16]);
    if (glpIntParam[16] < 0 || glpIntParam[16] > 1)
    {
        error ("glpk: the parameter \"presol\" must be 0 (do NOT use LP presolver [default]) or 1 (use LP presolver [default])");
        return retval;
    }

    //-- Generating cuts
    OCTAVE_GLPK_GET_INT_PARAM ("usecuts", glpIntParam[17]);
    if (glpIntParam[17] < 0 || glpIntParam[17] > 5)
    {
        error ("glpk: the parameter \"usecuts\" must be 0 (do NOT generate cuts) or 1 (generate Gomory's cuts [default]) or 2 (generate mixed integer rounding cuts) or 3 (generate cover cuts) or 4 (generate clique cuts) or 5 (generate all cuts)");
        return retval;
    }

    //-- PreProcessing
    OCTAVE_GLPK_GET_INT_PARAM ("pprocess", glpIntParam[18]);
    if (glpIntParam[18] < 0 || glpIntParam[18] > 2)
    {
        error ("glpk: the parameter \"pprocess\" must be 0 (disable preprocessing) or 1 (preprocessing for root only) or 2 (preprocessing for all levels [default])");
        return retval;
    }

    //-- Binarize
    OCTAVE_GLPK_GET_INT_PARAM ("binarize", glpIntParam[19]);
    if (glpIntParam[19] < 0 || glpIntParam[19] > 1)
    {
        error ("glpk: the parameter \"binarize\" must be 0 (do use binarization [default]) or 1 (replace general integer variable by binary ones)");
        return retval;
    }

    //-- LPsolver option
    volatile int lpsolver = 1;
    OCTAVE_GLPK_GET_INT_PARAM ("lpsolver", lpsolver);
    if (lpsolver < 1 || lpsolver > 3)
    {
        error ("glpk: the parameter \"lpsolver\" must be 1 (simplex method [default]) or 2 (interior point method) or 3 (LP in exact arithmetic)");
        return retval;
    }

    //-- Save option
    volatile int save_pb = 0;
    std::string filename("outpb");
    std::string filetype("cplex");
    OCTAVE_GLPK_GET_INT_PARAM ("save", save_pb);
    save_pb = save_pb != 0;
    if (save_pb) {
        octave_value name = PARAM.getfield("savefilename");

        if (name.is_defined() )
        {
            if (!name.is_empty())
            {
                filename = name.string_value();
            }
        }

        octave_value type = PARAM.getfield("savefiletype");

        if(type.is_defined())
        {
            if (!name.is_empty())
            {
                filetype = type.string_value();
            }
        }

        if(filetype.compare("cplex") == 0)
        {
            filename += ".lp";
        }
        else
        {
            if(filetype.compare("fixedmps") == 0 || filetype.compare("freemps") == 0)
            {
                filename += std::string(".mps");
            }
            else
            {
                filename += std::string(".txt");
            }
        }
    }

    // MPS parameters
    //-- mpsinfo
    OCTAVE_GLPK_GET_INT_PARAM ("mpsinfo", glpIntParam[8]);
    //-- mpsobj
    OCTAVE_GLPK_GET_INT_PARAM ("mpsobj", glpIntParam[9]);
    if (glpIntParam[9] < 0 || glpIntParam[9] > 2)
    {
        error("glpk: the parameter \"mpsobj\" must be 0 (never output objective function row) or 1 (always output objective function row ) or 2 (output objective function row if the problem has no free rows [default])");
        return retval;
    }
    //-- mpsorig
    OCTAVE_GLPK_GET_INT_PARAM ("mpsorig", glpIntParam[10]);
    //-- mpswide
    OCTAVE_GLPK_GET_INT_PARAM ("mpswide", glpIntParam[11]);
    //-- mpsfree
    OCTAVE_GLPK_GET_INT_PARAM ("mpsfree", glpIntParam[12]);

    //-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    //-- Real parameters
    //-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    //-- Ratio test option
    OCTAVE_GLPK_GET_REAL_PARAM ("relax", 0);

    //-- Relative tolerance used to check if the current basic solution
    //-- is primal feasible
    OCTAVE_GLPK_GET_REAL_PARAM ("tolbnd", 1);

    //-- Absolute tolerance used to check if the current basic solution
    //-- is dual feasible
    OCTAVE_GLPK_GET_REAL_PARAM ("toldj", 2);

    //-- Relative tolerance used to choose eligible pivotal elements of
    //--  the simplex table in the ratio test
    OCTAVE_GLPK_GET_REAL_PARAM ("tolpiv", 3);

    OCTAVE_GLPK_GET_REAL_PARAM ("objll", 4);

    OCTAVE_GLPK_GET_REAL_PARAM ("objul", 5);

    OCTAVE_GLPK_GET_REAL_PARAM ("tmlim", 6);

    OCTAVE_GLPK_GET_REAL_PARAM ("outdly", 7);

    OCTAVE_GLPK_GET_REAL_PARAM ("tolint", 8);

    OCTAVE_GLPK_GET_REAL_PARAM ("tolobj", 9);

    OCTAVE_GLPK_GET_REAL_PARAM ("mipgap", 10);

    //-- Assign pointers to the output parameters
    ColumnVector xmin (mrowsc, octave_NA);
    double fmin = octave_NA;
    double status;
    ColumnVector lambda (mrowsA, octave_NA);
    ColumnVector redcosts (mrowsc, octave_NA);
    double time;
    double mem;

    int jmpret = setjmp (mark);

    if (jmpret == 0)
        glpk (sense, mrowsc, mrowsA, c, nz, rn.fortran_vec (),
              cn.fortran_vec (), a.fortran_vec (), b, ctype,
              freeLB.fortran_vec (), lb, freeUB.fortran_vec (), ub,
              vartype.fortran_vec (), isMIP, lpsolver, save_pb, filename.c_str(), filetype.c_str(),
              xmin.fortran_vec (), &fmin, &status, lambda.fortran_vec (),
              redcosts.fortran_vec (), &time, &mem);

    octave_scalar_map extra;

    if (! isMIP)
    {
        extra.assign ("lambda", lambda);
        extra.assign ("redcosts", redcosts);
    }

    extra.assign ("time", time);
    extra.assign ("mem", mem);

    retval(3) = extra;
    retval(2) = status;
    retval(1) = fmin;
    retval(0) = xmin;

#else

    gripe_not_supported ("glpk");

#endif

    return retval;
}
